# copyright 2014-2020 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <https://www.gnu.org/licenses/>.

"""A CWProxy class wraps a CubicWeb repository.

>>> import cwproxy
>>> p = cwproxy.CWProxy('https://www.cubicweb.org')
>>> a = p.rql('Any X,T WHERE X is Project, X title T')
>>> print(a.json())
"""

import sys
import json
import warnings
import requests
import hmac
import hashlib
from time import time, sleep
from datetime import datetime, date
from urllib import parse as urlparse
import ssl
from typing import List

from .builders import build_trinfo

if not getattr(ssl, "HAS_SNI", False):
    try:
        import urllib3.contrib.pyopenssl

        urllib3.contrib.pyopenssl.inject_into_urllib3()
    except ImportError:
        pass

RQLIO_API = "1.0"


class SignedRequestAuth(requests.auth.AuthBase):
    """Auth implementation for CubicWeb with cube signedrequest"""

    hash_algorithm = "SHA512"

    def __init__(self, token_id, secret):
        self.token_id = token_id
        self.secret = secret

    def get_headers_to_sign(self):
        return ("Content-%s" % self.hash_algorithm.upper(), "Content-Type", "Date")

    def __call__(self, req):
        content = b""
        if req.body:
            content = req.body
        if isinstance(content, str):
            content = content.encode("utf-8")
        req.headers["Content-%s" % self.hash_algorithm.upper()] = getattr(
            hashlib, self.hash_algorithm.lower()
        )(content).hexdigest()
        content_to_sign = (
            req.method
            + req.url
            + "".join(req.headers.get(field, "") for field in self.get_headers_to_sign())
        )
        content_signed = hmac.new(
            self.secret.encode("utf-8"),
            content_to_sign.encode("utf-8"),
            digestmod=self.hash_algorithm.lower(),
        ).hexdigest()
        req.headers["Authorization"] = "Cubicweb %s:%s" % (
            self.token_id,
            content_signed,
        )
        return req


class MD5SignedRequestAuth(SignedRequestAuth):
    """
    Like SignedRequestAuth except it signed its requests with MD5

    This is INSECURE, DON'T USED IT except for compatibility reasons
    """

    hash_algorithm = "MD5"

    def __init__(self, token_id, secret):
        super(MD5SignedRequestAuth, self).__init__(token_id, secret)
        warning_message = (
            "WARNING: you are using an INSECURE SIGNING HASH algorithm (md5), please move to "
            "sha512 by using the SignedRequestAuth instead of the MD5SignedRequestAuth. The "
            "MD5SignedRequestAuth class will be removed in the future"
        )
        sys.stderr.write(warning_message + "\n")
        warnings.warn(warning_message, DeprecationWarning)


def date_header_value() -> str:
    return datetime.utcnow().strftime("%a, %d %b %Y %H:%M:%S GMT")


class RemoteValidationError(Exception):
    pass


class NoResultError(Exception):
    "CWProxy.find*() called but result set is empty"


class NoUniqueEntity(Exception):
    "CWProxy.find_one() called but result set contains more than one entity"


class TaskFailedError(Exception):
    pass


class CWProxy(object):
    """CWProxy: A simple helper class to ease building CubicWeb_
        clients. It allows to:

        * execute RQL_ queries remotely (using rqlcontroller_),
        * access instances that requires authentication (using signedrequest_).

    .. _CubicWeb: http://www.cubicweb.org/
    .. _RQL: http://docs.cubicweb.org/annexes/rql/language
    .. _rqlcontroller: http://www.cubicweb.org/project/cubicweb-rqlcontroller/
    .. _signedrequest: http://www.cubicweb.org/project/cubicweb-signedrequest/
    """

    def __init__(self, base_url: str, auth=None, verify=None, timeout=None):
        """Create a CWProxy connection object to the :base_url: cubicweb app.

        :param auth: can be provided to handle authentication. For a
          cubicweb application providing the signedrequest_ feature,
          one can use the SignedRequestAuth authentifier.

        :param verify: can be False to disable server certificate
          checking, or the path to a CA bundle file.

        """
        purl = urlparse.urlparse(base_url)
        # we do **not** want urls to be built with a double /, (e.g.
        # http://host// or http://host//basepath)
        path = purl.path.strip("/")
        if path:
            path = "{:s}/{:s}".format(purl.netloc, path)
        else:
            path = purl.netloc
        self.base_url = (purl.scheme, path)
        self.auth = auth
        self.timeout = timeout
        self._ssl_verify = verify
        self._default_vid = "jsonexport"  # OR 'ejsonexport'?
        self._base_url = urlparse.urlunparse(self.base_url + ("", None, None, None))

    def handle_request(self, method: str, path: str, **kwargs):
        """Construct a requests.Request and send it through this proxy.

        Arguments are that of requests.request() except for `path` which will
        be used to build a full URL using the base URL bound to this proxy.
        """
        msg = "handle_request() got unexpected keyword argument '{}'"
        for unexpected in ("url", "auth"):
            if unexpected in kwargs:
                raise TypeError(msg.format(unexpected))
        url = self.build_url(path)
        kwargs["auth"] = self.auth
        kwargs.setdefault("headers", {}).update({"Date": date_header_value()})
        kwargs.setdefault("verify", self._ssl_verify)
        kwargs.setdefault("timeout", self.timeout)
        return requests.request(method, url, **kwargs)

    def build_url(self, path: str, query=None):
        """Build the URL to query from self.base_url and the given path

        :param path: can be a string or an iterable of strings; if it
            is a string, it can be a simple path, in which case the
            URL will be built from self.base_url + path, or it can be
            an "absolute URL", in which case it will be queried as is
            (the query argument is then ignored)

        :param query: can be a sequence of two-elements **tuples** or
            a dictionary (ignored if path is an absolute URL)

        """
        if query:
            query = urlparse.urlencode(query, doseq=True)
        if isinstance(path, (list, tuple)):
            path = "/".join(path)
        if path.startswith(self._base_url):
            assert query is None
            return path
        return urlparse.urlunparse(self.base_url + (path, None, query, None))

    def get(self, path: str, query=None):
        """Perform a GET on the cubicweb instance

        :param path: the path part of the URL that will be GET
        :param query: can be a sequence of two-element tuples or a doctionnary
        """
        headers = {"Accept": "application/json"}
        return self.handle_request("GET", path, params=query, headers=headers)

    def post(self, path: str, **data):
        """Perform a POST on the cubicweb instance

        :param path: the path part of the URL that will be GET
        :param **data: will be passed as the 'data' of the request
        """
        kwargs = {
            "headers": {"Accept": "application/json"},
            "data": data,
        }
        if "files" in data:
            kwargs["files"] = data.pop("files")
        return self.handle_request("POST", path, **kwargs)

    def post_json(self, path: str, payload):
        """Perform a POST on the cubicweb instance with application/json
        Content-Type.

        :param path: the path part of the URL that will be GET
        :param payload: native data to be sent as JSON (not encoded)
        """
        kwargs = {
            "headers": {"Accept": "application/json"},
            "json": payload,
        }
        return self.handle_request("POST", path, **kwargs)

    def view(self, vid: str, **args):
        """Perform a GET on <base_url>/view with <vid> and <args>

        :param vid: the vid of the page to retrieve
        :param **args: will be used to build the query string of the URL
        """
        args["vid"] = vid
        return self.get("/view", args)

    def execute(self, rql: str, args=None):
        """CW connection's like execute method.

        :param rql: should be a unicode string or a plain ascii string
        :param args: are the optional parameters used in the query (dict)
        """
        result = self.rqlio([(rql, args or {})])

        try:
            return result.json()[0]
        except json.decoder.JSONDecodeError as e:
            raise Exception(
                "Failed to code response as json. Response "
                "(code %s) content:\n> %s\n\nException: %s"
                % (result.status_code, result.content, e)
            )

        return result.json()[0]

    def rql(self, rql: str, path="view", **data):
        """Perform an urlencoded POST to /<path> with rql=<rql>

        :param rql: should be a unicode string or a plain ascii string
        (warning, no string formating is performed)
        :param path: the path part of the generated URL
        :param **data: the 'data' of the request
        """
        if rql.split()[0] in ("INSERT", "SET", "DELETE"):
            raise ValueError(
                "You must use the rqlio() method to make " "write RQL queries"
            )

        if not data.get("vid"):
            data["vid"] = self._default_vid
        if path == "view":
            data.setdefault("fallbackvid", "404")
        if rql:  # XXX may be empty?
            if not rql.lstrip().startswith("rql:"):
                # add the 'rql:' prefix to ensure given rql is considered has
                # plain RQL so CubicWeb won't attempt other interpretation
                # (e.g. eid, 2 or 3 word queries, plain text)
                rql = "rql:" + rql
            data["rql"] = rql

        headers = {
            "Accept": "application/json",
            "Date": date_header_value(),
        }
        params = {
            "url": self.build_url(path),
            "headers": headers,
            "verify": self._ssl_verify,
            "auth": self.auth,
            "data": data,
        }
        return requests.post(**params)

    def rqlio(self, queries):
        """Multiple RQL for reading/writing data from/to a CW instance.

        :param queries: list of queries, each query being a couple (rql, args)

        Example::

          queries = [('INSERT CWUser U: U login %(login)s, U upassword %(pw)s',
                      {'login': 'babar', 'pw': 'cubicweb rulez & 42'}),
                     ('INSERT CWGroup G: G name %(name)s',
                      {'name': 'pachyderms'}),
                     ('SET U in_group G WHERE G eid %(g)s, U eid %(u)s',
                      {'u': '__r0', 'g': '__r1'}),
                     ('INSERT File F: F data %(content)s, F data_name %(fn)s',
                      {'content': BytesIO('some binary data'),
                       'fn': 'toto.bin'}),
                    ]
          self.rqlio(queries)

        """
        headers = {
            "Accept": "application/json",
            "Date": date_header_value(),
        }
        files = self.preprocess_queries(queries)

        params = {
            "url": self.build_url(("rqlio", RQLIO_API)),
            "headers": headers,
            "verify": self._ssl_verify,
            "auth": self.auth,
            "files": files,
        }
        posted = requests.post(**params)
        if posted.status_code == 500:
            try:
                cause = posted.json()
            except Exception as exc:
                raise RemoteValidationError("%s (%s)", exc, posted.text)
            else:
                if "reason" in cause:
                    # was a RemoteCallFailed
                    raise RemoteValidationError(cause["reason"])
        return posted

    def preprocess_queries(self, queries):
        """Pre process queries arguments to replace binary content by
        files to be inserted in the multipart HTTP query

        :param queries: list of queries, each query being a couple (rql, args)

        Any value that have a read() method will be threated as
        'binary content'.

        In the RQL query, binary value are replaced by unique '__f<N>'
        references (the ref of the file object in the multipart HTTP
        request).
        """

        files = {}
        for query_idx, (rql, args) in enumerate(queries):
            if args is None:
                continue
            for arg_idx, (k, v) in enumerate(args.items()):
                if hasattr(v, "read") and callable(v.read):
                    # file-like object
                    fid = args[k] = "__f%d-%d" % (query_idx, arg_idx)
                    files[fid] = v
                elif isinstance(v, (date, datetime)):
                    args[k] = v.isoformat()
        files["json"] = ("json", json.dumps(queries), "application/json")
        return files

    def _set_rql_request(self, rql_request: str, kwargs, sep=",") -> str:
        args = [
            "X {property_name:s} %({property_name:s})s".format(
                property_name=property_name
            )
            for property_name in kwargs
        ]
        if args:
            rql_request = u"{:s}{:s} {:s}".format(rql_request, sep, ", ".join(args))
        return rql_request

    def _rql_args_query(self, rql_request: str, kwargs, sep=",") -> list:
        rql_request = self._set_rql_request(rql_request, kwargs, sep)
        response = self.rqlio([(rql_request, kwargs)])
        response.raise_for_status()
        results = response.json()
        return [row[0] for row in results[0]]

    def count(self, entity_type: str, **kwargs) -> int:
        """Return number of entities with the given type and
        properties in a CW instance.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values
        :return: number of entities
        :rtype: int

        Example::

          >>> self.count('CWUser')
          3
          >>> self.count('CWUser', login='rms')
          1

        """
        rql_query = "Any COUNT("
        rql_query += "1" if kwargs else "X"
        rql_query += ") WHERE X is {:s}".format(entity_type)
        return int(self._rql_args_query(rql_query, kwargs)[0])

    def exist(self, entity_type: str, **kwargs) -> bool:
        """Return true if there is at least one entity with the given type and
        properties in a CW instance and false otherwise.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values
        :return: whether such an entity exists
        :rtype: boolean

        Example::

          >>> self.exist('CWUser', login='toto')
          False

        """
        rql_query = "Any "
        rql_query += "1" if kwargs else "X"
        rql_query += " LIMIT 1 WHERE X is {:s}".format(entity_type)
        try:
            return bool(self._rql_args_query(rql_query, kwargs))
        except RemoteValidationError as e:
            if "unknown entity type {:s}".format(entity_type) in str(e):
                return False
            raise e

    def find(self, entity_type: str, **kwargs) -> List[int]:
        """Return eid(s) of entitie(s) with the given type and properties in a
        CW instance.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values
        :return: list of eid(s)
        :rtype: list

        Example::

          >>> self.find('CWUser')
          [20, 21]
          >>> self.find('CWUser', login='admin')
          [20]
        """
        return self._rql_args_query("Any X WHERE X is {:s}".format(entity_type), kwargs)

    def find_one(self, entity_type: str, **kwargs) -> int:
        """Return eid of the unique entity with the given type and properties
        in a CW instance. If there is none or multiple, it throws an exception
        that indicates the problem (NoResultError or NoUniqueEntity).

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values
        :return: eid
        :rtype: integer
        :raises NoResultError: there was no matching entity
        :raises NoUniqueEntity: there was multiple matching entities

        Example::

          >>> self.find_one('CWUser', login='admin')
          20
        """
        eids = self._rql_args_query(
            "Any X LIMIT 2 WHERE X is {:s}".format(entity_type), kwargs
        )
        if len(eids) == 0:
            raise NoResultError("No result for {:s}".format(entity_type))
        if len(eids) > 1:
            raise NoUniqueEntity("Cannot find unique {:s}".format(entity_type))
        return eids[0]

    def find_last_created(self, entity_type: str, **kwargs) -> int:
        """Return eid of the last created entity with the given type and
        properties in a CW instance. If there is none, it throws an exception
        that indicates the problem (NoResultError).

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values
        :return: eid
        :rtype: integer
        :raises NoResultError: there was no matching entity

        Example::

          >>> self.find_last_created('CWUser')
          20
          >>> self.find_last_created('Blog', title='MyBlog')
          22
        """
        eid = self._rql_args_query(
            "Any X ORDERBY D DESC LIMIT 1 "
            "WHERE X is {:s}, X creation_date D".format(entity_type),
            kwargs,
        )
        if not eid:
            raise NoResultError("No result for {:s}".format(entity_type))
        return eid[0]

    def get_state(self, eid: int) -> str:
        """Return name of the state of the entity with the given eid or nothing
        if the eid does not exist or if the entity with the given eid has no
        state.

        :param eid: eid of an entity
        :return: the name of the state (if one)
        :rtype: str

        Example::

          >>> self.get_state(1001)
          'wfs_finished'
        """
        response = self.rqlio(
            [
                (
                    "Any SN LIMIT 1 WHERE E eid %(eid)s, E in_state S, S name SN",
                    {"eid": eid},
                )
            ]
        )
        response.raise_for_status()
        rset = response.json()
        if rset and rset[0] and rset[0][0]:
            return rset[0][0][0]
        return None

    def wait_for_status(self, eid: int, status, timeout=60, timesleep=1) -> None:
        """Wait that the entity with given eid to be in status given. If it is
        not the case after timeout, a related exception is raised. The state is
        fetched and checked, then it sleeps if it is not yet the status given.

        :param eid: eid of an entity with a state
        :param status: status to wait for
        :param timeout: maximum time to wait for given status
        :param timesleep: time between each status fetch
        :raises TaskFailedError: status given was not failed and it failed
        :raises TimeoutError: timeout has expired

        Example::

          >>> self.wait_for_status(30, 'wfs_finished')
        """
        start_time = int(time())
        while True:
            sleep(timesleep)
            current_status = self.get_state(eid)
            if current_status == status:
                break
            if current_status == "wfs_failed":
                raise TaskFailedError(eid)
            if int(time()) - start_time >= timeout:
                raise TimeoutError(eid)

    def wait_for_finish(self, eid: int, *args, **kwargs) -> None:
        """Wait that the entity with given eid to be in status finished. If it
        is not the case after timeout, a related exception is raised. The state
        is fetched and checked, then it sleeps if it is not yet in the status
        finished.

        :param eid: eid of an entity with a state
        :param timeout: maximum time to wait for given status
        :param timesleep: time between each status fetch
        :raises TaskFailedError: the task failed according to state
        :raises TimeoutError: timeout has expired

        Example::

          >>> self.wait_for_finish(30)
        """
        return self.wait_for_status(eid, "wfs_finished", *args, **kwargs)

    def change_state(self, eid: int, status: str):
        """Try to change the state with the one given for the entity that have
        the given eid.

        :param eid: eid of an entity with a state
        :param status: new status to set

        Example::

          >>> self.change_state(30, 'wft_start')
        """
        return self.rqlio([build_trinfo(eid, status)])

    def insert(self, entity_type: str, **kwargs):
        """Insert an entity of the given type with given values of attributes.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values

        Example::

          >>> self.insert('Project', name='start-up nation', author='Macron')
        """
        return self._rql_args_query("INSERT {:s} X: ".format(entity_type), kwargs, "")

    def insert_if_not_exist(self, entity_type, **kwargs):
        """Insert an entity of the given type with given values of attributes
        if it does not already exist.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values

        Example::

          >>> self.insert_if_not_exist('Project', name='cwclientlib')
        """
        if not self.exist(entity_type, **kwargs):
            return self.insert(entity_type, **kwargs)

    def delete(self, entity_type: str, **kwargs):
        """Delete all entities that have the given type
        and with given values of attributes.

        :param entity_type: entity type name
        :param kwargs: list of properties with associated values

        Example::

          >>> self.delete('Project', name='start-up nation', author='Macron')
        """
        return self._rql_args_query(
            "DELETE {:s} X WHERE ".format(entity_type), kwargs, ""
        )



# -*- coding: utf-8 -*-
#
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

"""provide RQL autocompletion for statements starting with::

   >>> whatever.execute('Any X WHERE X is
   >>> rql('Any X WHERE X is

"""
from __future__ import print_function

import re
import readline
from rlcompleter import Completer
import requests

from cwclientlib.cwproxy import date_header_value


class AbstractMatcher(object):
    """Abstract class for CWShellCompleter's matchers.

    A matcher should implement a ``possible_matches`` method. This
    method has to return the list of possible completions for user's input.
    Because of the python / readline interaction, each completion should
    be a superset of the user's input.

    NOTE: readline tokenizes user's input and only passes last token to
    completers.
    """

    def possible_matches(self, text):
        """return possible completions for user's input.

        Parameters:
            text: the user's input

        Return:
            a list of completions. Each completion includes the original input.
        """
        raise NotImplementedError()


class RQLExecuteMatcher(AbstractMatcher):
    """Custom matcher for rql queries.

    If user's input starts with ``rql(`` or ``session.execute(`` and
    the corresponding rql query is incomplete, suggest some valid completions.
    """

    query_match_rgx = re.compile(
        r"(?P<func_prefix>\s*(?:rql)"  # match rql, possibly indented
        r"|"  # or
        r"\s*(?:\w+\.execute))"  # match .execute, possibly indented
        # end of <func_prefix>
        r"\("  # followed by a parenthesis
        r'(?P<quote_delim>["\'])'  # a quote or double quote
        r"(?P<parameters>.*)"
    )  # and some content

    def __init__(self, client):
        self.client = client

    @staticmethod
    def match(text):
        """check if ``text`` looks like a call to ``rql`` or ``session.execute``

        Parameters:
            text: the user's input

        Returns:
            None if it doesn't match, the query structure otherwise.
        """
        query_match = RQLExecuteMatcher.query_match_rgx.match(text)
        if query_match is None:
            return None
        parameters_text = query_match.group("parameters")
        quote_delim = query_match.group("quote_delim")
        # first parameter is fully specified, no completion needed
        if re.match(r"(.*?)%s" % quote_delim, parameters_text) is not None:
            return None
        func_prefix = query_match.group("func_prefix")
        return {
            # user's input
            "text": text,
            # rql( or session.execute(
            "func_prefix": func_prefix,
            # offset of rql query
            "rql_offset": len(func_prefix) + 2,
            # incomplete rql query
            "rql_query": parameters_text,
        }

    def possible_matches(self, text):
        """call ``rql.suggestions`` component to complete user's input."""
        # readline will only send last token, but we need the entire
        # user's input
        user_input = readline.get_line_buffer()
        query_struct = self.match(user_input)
        if query_struct is None:
            return []
        else:
            # we must only send completions of the last token =>
            # compute where it starts relatively to the rql query
            # itself.
            compl_offset = readline.get_begidx() - query_struct["rql_offset"]
            rql_query = query_struct["rql_query"]
            return [
                suggestion[compl_offset:]
                for suggestion in get_suggestions(self.client, rql_query)
            ]


class DefaultMatcher(AbstractMatcher):
    """Default matcher: delegate to standard's `rlcompleter.Completer`` class"""

    def __init__(self, namespace=None):
        self.completer = Completer(namespace)

    def possible_matches(self, text):
        if "." in text:
            return self.completer.attr_matches(text)
        else:
            return self.completer.global_matches(text)


class CWShellCompleter(object):
    """Custom auto-completion helper for cubicweb-ctl shell.

    ``CWShellCompleter`` provides a ``complete`` method suitable for
    ``readline.set_completer``.

    Attributes:
        matchers: the list of ``AbstractMatcher`` instances that will suggest
                  possible completions

    The completion process is the following:

    - readline calls the ``complete`` method with user's input,
    - the ``complete`` method asks for each known matchers if
      it can suggest completions for user's input.
    """

    def __init__(self, client, namespace=None):
        # list of matchers to ask for possible matches on completion
        self.matchers = [DefaultMatcher(namespace)]
        self.matchers.insert(0, RQLExecuteMatcher(client))

    def complete(self, text, state):
        """readline's completer method

        see:
          http://docs.python.org/2/library/readline.html#readline.set_completer
        for more details.

        Implementation inspired by `rlcompleter.Completer`
        """
        if state == 0:
            # reset self.matches
            self.matches = []
            for matcher in self.matchers:
                matches = matcher.possible_matches(text)
                if matches:
                    self.matches = matches
                    break
            else:
                return None  # no matcher able to handle `text`
        try:
            return self.matches[state]
        except IndexError:
            return None


def get_suggestions(client, user_input):
    try:
        data = {
            "fname": "rql_suggest",
            "term": user_input,
        }
        headers = {
            "Accept": "application/json",
            "Data": date_header_value(),
        }
        sugg = requests.post(
            url=client.build_url("/json"),
            headers=headers,
            verify=client._ssl_verify,
            auth=client.auth,
            data=data,
        ).json()
    except Exception as exc:
        print("Failed to fetch suggestions: {0}".format(exc))
    return sugg


def setup_autocompleter(client, namespace=None):
    import readline
    from .cwcompleter import CWShellCompleter

    completer = CWShellCompleter(client, namespace)
    readline.set_completer(completer.complete)
    readline.parse_and_bind("tab: complete")



# -*- coding: utf-8 -*-
#
# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

import os

from .cwproxy import CWProxy, SignedRequestAuth, MD5SignedRequestAuth


def cwproxy_for(instance, **kw):
    """Build a CWProxy instance for a given cubicweb instance, reading
    authentication credentials from a ini, a json config file or a
    YAML file.

    :param instance: can be either the name of the instance in the config
    file, or its URL

    :param **kw: will be passed to the proxy class constructor; the
    instanciated proxy class can be specified as a proxycls named
    argument

    The config file is ~/.config/cwclientlibrc by default (ini
    format), but this can be changed by setting the CWCLCONF
    environment variable (the ~/.config can be overriden by the
    XDG_CONFIG_HOME env var, see
    http://standards.freedesktop.org/basedir-spec/basedir-spec-0.6.html).

    By default, the configuration format is the ini file format ; it
    must look like:

      [cwo]
      url = https://www.cubicweb.org/
      token-id = my token id
      secret = <my secret>

    If the file name ends with .json, it will be read by a JSON
    parser, like:

      {'cwo': {'url': 'https://www.cubicweb.org/',
               'token-id': 'my token id',
               'secret': '<my secret>'},
      }

    If the file name ends with .yaml, it will be read by a YAML
    parser, like:

      cwo:
        url: https://www.cubicweb.org/
        token-id: my token id
        secret: <my secret>

    Supported authentications are 'signedrequest'. The
    default authentication mechanism is 'signedrequest'

    """
    cfg = get_config()
    # if instance is not the id of a cw endpoint, loop in config
    # entries to match on url
    if instance not in cfg:
        instances = [
            inst
            for inst, instcfg in cfg.items()
            if instance.startswith(instcfg.get("url"))
        ]
        if len(instances) == 1:
            instance = instances[0]
    if instance not in cfg:
        raise ValueError("Cannot find a configuration entry for %r" % instance)

    cfg = cfg[instance]
    auth_mech = cfg.get("auth-mech", "signedrequest")
    url = cfg.get("url")
    if not url:
        raise ValueError('Missing "url" configuration option')

    if auth_mech in ("signedrequest", "md5signedrequest"):
        tokenid = cfg.get("token-id")
        secret = cfg.get("secret")
        if not tokenid or not secret:
            raise ValueError(
                'Missing "token-id" or "secret" configuration '
                "option for signedrequest"
            )
        if auth_mech == "signedrequest":
            auth = SignedRequestAuth(tokenid, secret)
        elif auth_mech == "md5signedrequest":
            auth = MD5SignedRequestAuth(tokenid, secret)

    else:
        raise ValueError(
            "Unknown authentication mechanism (auth-mech): " "%r" % auth_mech
        )
    if "server-ca" in cfg:
        kw["verify"] = cfg["server-ca"]

    proxy_cls = kw.pop("proxycls", CWProxy)
    return proxy_cls(url, auth, **kw)


def get_config():
    defaultcfg = os.path.join(
        os.environ.get("XDG_CONFIG_HOME", "~/.config"), "cwclientlibrc"
    )
    cfgfile = os.path.expanduser(os.environ.get("CWCLCONF", defaultcfg))
    if not os.path.isfile(cfgfile):
        raise EnvironmentError("Cannot find the configuration file %r" % cfgfile)
    if os.name == "posix":
        perms = os.stat(cfgfile).st_mode % (0o1000)
        if (perms & 0o600) != perms:
            raise EnvironmentError(
                "%r permissions should be 0600 or 0400 (%o)" % (cfgfile, perms)
            )

    if cfgfile.endswith(".json"):
        import json

        with open(cfgfile) as f:
            cfg = json.load(f)
    elif cfgfile.endswith(".yaml"):
        import yaml

        with open(cfgfile) as f:
            cfg = yaml.load(f)
    else:
        from configparser import ConfigParser

        cp = ConfigParser()
        cp.read(cfgfile)
        cfg = {section: dict(cp.items(section)) for section in cp.sections()}
    return cfg



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function
import sys

from . import cwproxy_for, get_config
from .cwproxy import CWProxy, RemoteValidationError

import argparse

try:
    import argcomplete
except ImportError:
    argcomplete = None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "endpoint", nargs=1, choices=sorted(get_config()), help="endpoint"
    )
    parser.add_argument("query", nargs=1, help="RQL query")
    parser.add_argument(
        "-V",
        "--vid",
        dest="vid",
        default=None,
        help="The vid to use (defaults to ejsonexport)",
    )
    parser.add_argument(
        "-c",
        "--server-ca",
        dest="ca",
        default=None,
        help="Bundle CA to use to verify server certificate",
    )
    parser.add_argument(
        "-S",
        "--no-ssl",
        dest="verify",
        default=True,
        action="store_false",
        help=("do NOT verify ssl server certificate; " "ignored if --ca is given"),
    )

    if argcomplete is not None:
        argcomplete.autocomplete(parser)
    args = parser.parse_args()

    url = args.endpoint[0]
    verify = args.ca or args.verify  # path to the bundle CA or bool
    try:
        client = cwproxy_for(url, verify=verify)
    except ValueError:
        if url.startswith(("http://", "https://")):
            client = CWProxy(url, verify=verify)
        else:
            raise
    rql = args.query[0]
    try:
        resp = client.rql(rql, vid=args.vid)
        try:
            print(resp.json())
        except ValueError:
            print(resp.text.encode("utf-8"))
    except RemoteValidationError as exc:
        print("** Request failed: {0}".format(exc), file=sys.stderr)
        print("** RQL was %r" % (rql,), file=sys.stderr)


if __name__ == "__main__":
    main()



# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

import os
import tempfile
from unittest import TestCase
import json
import yaml
from contextlib import contextmanager

from cwclientlib import cwproxy_for


CFG = {
    "sr ok": {
        "url": "http://www.cubicweb.org",
        "auth-mech": "signedrequest",
        "token-id": "toto",
        "secret": "mysecret",
    },
    "sr2 ok": {
        "url": "http://www.cubicweb.org",
        "token-id": "toto",
        "secret": "mysecret",
    },
    "sr err": {
        "token-id": "toto",
        "secret": "mysecret",
    },
    "sr2 err": {
        "url": "http://www.cubicweb.org",
        "secret": "mysecret",
    },
    "sr3 err": {
        "url": "http://www.cubicweb.org",
        "token-id": "toto",
    },
    "auth-mech err": {
        "url": "http://www.cubicweb.org",
        "auth-mech": "other mech",
    },
}


def cfg2ini(cfg):
    cfg = (
        "[%s]\n%s" % (s, "".join("%s = %s\n" % (k, v) for k, v in sval.items()))
        for s, sval in cfg.items()
    )
    return "\n".join(cfg)


class CWProxyForTC(TestCase):
    """Basic tests for CWProxy"""

    @contextmanager
    def temp_config_file(self, suffix="rc", mode=0o600):
        with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as f:
            if suffix == ".json":
                f.write(json.dumps(CFG).encode("utf-8"))
            elif suffix == ".yaml":
                f.write(yaml.dump(CFG, default_flow_style=False).encode("utf-8"))
            else:
                f.write(cfg2ini(CFG).encode("utf-8"))
        os.chmod(f.name, mode)
        os.environ["CWCLCONF"] = f.name
        try:
            yield
        finally:
            os.remove(f.name)

    def test_signedrequest_ini(self):
        with self.temp_config_file():
            for src in CFG:
                if src.endswith("ok"):
                    self.assertTrue(cwproxy_for(src))
                else:
                    with self.assertRaises(ValueError):
                        cwproxy_for(src)

    def test_signedrequest_json(self):
        with self.temp_config_file(suffix=".json"):
            for src in CFG:
                if src.endswith("ok"):
                    self.assertTrue(cwproxy_for(src))
                else:
                    with self.assertRaises(ValueError):
                        cwproxy_for(src)

    def test_signedrequest_yaml(self):
        with self.temp_config_file(suffix=".yaml"):
            for src in CFG:
                if src.endswith("ok"):
                    self.assertTrue(cwproxy_for(src))
                else:
                    with self.assertRaises(ValueError):
                        cwproxy_for(src)

    def test_signedrequest_configfile_permissions(self):
        src = "sr ok"

        for mode in (0o400, 0o600):
            with self.temp_config_file(mode=mode):
                self.assertTrue(cwproxy_for(src))

        for mode in (0o700, 0o640, 0o604):
            with self.temp_config_file(mode=0o700):
                with self.assertRaises(EnvironmentError):
                    cwproxy_for(src)


if __name__ == "__main__":
    from unittest import main

    main()



# copyright 2014-2020 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <https://www.gnu.org/licenses/>.

from io import BytesIO
import json
from datetime import date, datetime, timedelta

import pytz

import unittest

from cwclientlib.cwproxy import (
    CWProxy,
    SignedRequestAuth,
    NoResultError,
    NoUniqueEntity,
)

from cubicweb.devtools.httptest import CubicWebServerTC


class CWProxyClientTC(CubicWebServerTC):
    """Basic tests for CWProxy"""

    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            token = cnx.create_entity(
                "AuthToken", id=u"testing", enabled=True, token_for_user=cnx.user.eid
            )
            self.admineid = cnx.user.eid
            self.token_id = token.id
            self.token_secret = token.token
            cnx.commit()

    def _addUser(self, login: str, password: str, group_name: str = "users"):
        queries = [
            (
                "INSERT CWUser U: U login %(login)s, U upassword %(pw)s, "
                "U in_group G WHERE G name %(group_name)s",
                {"login": login, "pw": password, "group_name": group_name},
            )
        ]
        results = self.client.rqlio(queries).json()
        self.assertTrue(results)
        eid = results[0][0][0]
        return eid

    @property
    def client(self):
        auth = SignedRequestAuth(token_id=self.token_id, secret=self.token_secret)
        base_url = self.config["base-url"]
        return CWProxy(base_url=base_url, auth=auth)

    def test_rql(self):
        self.assertEqual(
            [["admin"], ["anon"]], self.client.rql("Any L WHERE U login L").json()
        )

        self.assertEqual(
            [[self.admineid]], self.client.rql('Any U WHERE U login "admin"').json()
        )

    def test_execute(self):
        self.assertEqual(
            [[self.admineid]],
            self.client.execute("Any U WHERE U login %(l)s", {"l": "admin"}),
        )

    def test_rql_vid(self):
        admin = self.client.rql(
            'Any U WHERE U login "admin"', vid="ejsonexport"
        ).json()[0]
        self.assertIn("login", admin.keys())

    def test_rql_path(self):
        result = self.client.rql(
            'Any U WHERE U login "admin"', path="test-controller"
        ).text
        self.assertEqual("coucou", result)

    def test_bad_rql_path(self):
        result = self.client.rql('CWGroup U WHERE U login "admin"')
        self.assertEqual(result.status_code, 500)

    def test_rqlio(self):
        queries = [
            (
                "INSERT CWUser U: U login %(login)s, U upassword %(pw)s",
                {"login": "babar", "pw": "cubicweb rulez & 42"},
            ),
            ("INSERT CWGroup G: G name %(name)s", {"name": "pachyderms"}),
            (
                "SET U in_group G WHERE G eid %(g)s, U eid %(u)s",
                {"u": "__r0", "g": "__r1"},
            ),
        ]

        results = self.client.rqlio(queries).json()
        self.assertEqual(3, len(results))
        babar = results[0][0][0]
        pach = results[1][0][0]

        with self.admin_access.client_cnx() as cnx:
            users = cnx.find("CWUser", eid=babar)
            self.assertEqual(1, len(users))
            self.assertEqual("babar", users.one().login)

            groups = cnx.find("CWGroup", eid=pach)
            self.assertEqual(1, len(groups))
            self.assertEqual("pachyderms", groups.one().name)

        self.assertEqual([babar, pach], results[2][0])

    def test_rqlio_no_kwargs(self):
        queries = [
            (
                'INSERT CWUser U: U login "babar", '
                'U upassword "cubicweb rulez & 42", '
                'U in_group G WHERE G name "users"',
                None,
            ),
        ]
        results = self.client.rqlio(queries).json()
        self.assertEqual(1, len(results))
        babar = results[0][0][0]
        with self.admin_access.client_cnx() as cnx:
            users = cnx.find("CWUser", eid=babar)
            self.assertEqual(1, len(users))
            self.assertEqual("babar", users.one().login)

    def test_rqlio_multiple_binary_arguments(self):
        queries = [
            (
                'INSERT User X: X name "babar",'
                " X picture %(picture)s, X ssh_pubkey %(ssh_pubkey)s",
                {"picture": BytesIO(b"nice photo"), "ssh_pubkey": BytesIO(b"12345")},
            ),
        ]

        eid = self.client.rqlio(queries).json()[0][0][0]
        with self.admin_access.client_cnx() as cnx:
            user = cnx.entity_from_eid(eid)
            self.assertEqual(user.picture.getvalue(), b"nice photo")
            self.assertEqual(user.ssh_pubkey.getvalue(), b"12345")

    def test_rqlio_datetime(self):
        md = datetime.now(pytz.timezone("Europe/Paris")) - timedelta(days=1)
        queries = [
            (
                'INSERT CWUser U: U login "babar", '
                "U creation_date %(cd)s, "
                "U modification_date %(md)s, "
                'U upassword "cubicweb rulez & 42", '
                'U in_group G WHERE G name "users"',
                {"cd": md, "md": md},
            ),
        ]

        eid = self.client.rqlio(queries).json()[0][0][0]
        with self.admin_access.client_cnx() as cnx:

            user = cnx.entity_from_eid(eid)
            self.assertEqual(md, user.creation_date)
            self.assertEqual(
                md.tzinfo.utcoffset(md),
                user.creation_date.tzinfo.utcoffset(user.creation_date),
            )
            self.assertEqual(md, user.modification_date)

        # actually only check date type is handled, we can't expect a
        # given number of entities as it depends on the database cache
        # creation time
        res = self.client.execute(
            "CWUser X WHERE X creation_date < %(today)s", {"today": date.today()}
        )
        self.assertTrue(res)

    def test_exist(self):
        eid = self._addUser("babar", "cubicweb rulez & 42")

        self.assertTrue(self.client.exist("CWUser"))
        self.assertTrue(self.client.exist("CWUser", login="babar"))
        self.assertTrue(self.client.exist("CWUser", eid=eid))
        self.assertTrue(self.client.exist("CWUser", eid=eid, login="babar"))
        self.assertFalse(self.client.exist("CWNothing"))
        self.assertFalse(
            self.client.exist("CWUser", login="Sauron", upassword="DOMINATION")
        )

    def test_count(self):
        nb = self.client.count("CWUser")
        self._addUser("Richard Stallman", "GNU")
        self.assertEqual(self.client.count("CWUser"), nb + 1)
        self._addUser("Jimmy Wales", "Wikipédia")
        self.assertEqual(self.client.count("CWUser"), nb + 2)
        self.assertEqual(self.client.count("CWUser", login="Richard Stallman"), 1)

    def test_find(self):
        eid_babar = self._addUser("babar", "cubicweb rulez & 42")
        eid_sacha = self._addUser("sacha", "catch them all!")

        results = self.client.find("CWUser")
        self.assertTrue(results)
        self.assertIn(eid_babar, (row for row in results))
        self.assertIn(eid_sacha, (row for row in results))

        results = self.client.find("CWUser", login="babar")
        self.assertTrue(results)
        self.assertEqual(results[0], eid_babar)

        results = self.client.find("CWUser", eid=eid_babar, login="babar")
        self.assertTrue(results)
        self.assertEqual(results[0], eid_babar)

        self.assertFalse(self.client.find("CWUser", login="Bill Gates"))

    def test_find_one(self):
        eid_rms = self._addUser("Richard Stallman", "GNU")
        eid_jim = self._addUser("Jimmy Wales", "Wikipédia")

        self.assertEqual(
            self.client.find_one("CWUser", login="Richard Stallman"), eid_rms
        )
        self.assertEqual(self.client.find_one("CWUser", login="Jimmy Wales"), eid_jim)

        with self.assertRaises(NoResultError):
            self.client.find_one("CWUser", login="Bill Gates")

        with self.assertRaises(NoUniqueEntity):
            self.client.find_one("CWUser")

    def test_find_last_created(self):
        eid = self._addUser("Hurd", "GNU")
        self.assertEqual(self.client.find_last_created("CWUser"), eid)

        eid = self._addUser("Linux", "Linus")
        self.assertEqual(self.client.find_last_created("CWUser"), eid)

    def test_get_state(self):
        eid = self._addUser("Hurd", "GNU")
        self.assertEqual(self.client.get_state(eid), "activated")

    def test_wait_for_status(self):
        eid = self._addUser("Hurd", "GNU")
        self.client.wait_for_status(eid, "activated")

        with self.assertRaises(TimeoutError):
            self.client.wait_for_status(eid, "wfs_failed", 0, 0)

    def test_wait_for_finish(self):
        eid = self._addUser("Hurd", "GNU")
        with self.assertRaises(TimeoutError):
            self.client.wait_for_finish(eid, 0, 0)

    def test_handle_request(self):
        with self.admin_access.cnx() as cnx:
            eid = cnx.find("CWUser", login=u"admin")[0][0]
        response = self.client.handle_request(
            "GET",
            "cwuser/{}".format(eid),
            headers={"Accept": "application/custom+json"},
        )
        # XXX getting 403 (as in test_get/test_view), don't know why.
        # self.assertEqual(response.status_code, 200)
        request = response.request
        self.assertEqual(request.method, "GET")
        self.assertEqual(
            request.url, "{}cwuser/{}".format(self.config["base-url"], eid)
        )
        headers = request.headers
        self.assertEqual(headers["Accept"], "application/custom+json")
        self.assertIn("Date", headers)
        self.assertIn("Authorization", headers)
        self.assertIn("Content-SHA512", headers)
        self.assertIn("Cubicweb testing:", headers["Authorization"])

    def test_handle_request_bad_parameter(self):
        with self.assertRaisesRegex(TypeError, "got unexpected keyword argument 'url'"):
            self.client.handle_request("GET", 1, url="x")

    @unittest.skip("to be completed and make pass")
    def test_get(self):
        with self.admin_access.client_cnx() as cnx:
            eid = cnx.find("CWEType", name=u"CWUser")[0][0]
        response = self.client.get("/cwetype/{}".format(eid))
        self.assertEqual(response.status_code, 200)

    @unittest.skip("to be completed and make pass")
    def test_view(self):
        vid = "jsonexport"
        with self.admin_access.web_request() as req:
            rset = req.find("CWEType", name=u"CWUser")
            json_view = req.view(vid, rset=rset)
        response = self.client.view(vid, eid=rset[0][0])
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.json(), json_view[0])

    def test_post(self):
        files = [("myfile", BytesIO(b"some content"))]
        data = dict(a="hello", files=files)
        response = self.client.post("/foo/bar", **data)
        self.assertEqual(response.status_code, 404)
        request = response.request
        self.assertEqual(request.method, "POST")
        headers = request.headers
        content_type, _ = headers["Content-Type"].split("; ", 1)
        self.assertEqual(content_type, "multipart/form-data")
        bodylines = request.body.splitlines()
        self.assertIn(b'Content-Disposition: form-data; name="a"', bodylines)
        self.assertIn(b"hello", bodylines)
        self.assertIn(
            b'Content-Disposition: form-data; name="myfile"; ' b'filename="myfile"',
            bodylines,
        )
        self.assertIn(b"some content", bodylines)

    def test_post_json(self):
        data = {"foo": 1}
        response = self.client.post_json("/foo/bar", data)
        self.assertEqual(response.status_code, 404)
        request = response.request
        self.assertEqual(request.method, "POST")
        headers = request.headers
        content_type = headers["Content-Type"]
        self.assertEqual(content_type, "application/json")
        body = request.body
        self.assertEqual(json.loads(body.decode("utf-8")), data)

    def test_insert(self):
        self.assertFalse(self.client.find("EmailAddress", alias="CSE"))
        self.client.insert("EmailAddress", alias="CSE", address="dp@logilab.fr")
        self.assertTrue(self.client.find("EmailAddress", alias="CSE"))

    def test_insert_if_not_exist(self):
        self.assertFalse(self.client.find("EmailAddress", alias="CSE"))
        self.client.insert_if_not_exist(
            "EmailAddress", alias="CSE", address="dp@logilab.fr"
        )
        self.assertEqual(self.client.count("EmailAddress", alias="CSE"), 1)
        self.client.insert_if_not_exist(
            "EmailAddress", alias="CSE", address="dp@logilab.fr"
        )
        self.assertEqual(self.client.count("EmailAddress", alias="CSE"), 1)

    def test_delete(self):
        self.client.insert(
            "EmailAddress", alias="juliette", address="juliette@logilab.fr"
        )
        self.assertTrue(self.client.find("EmailAddress", alias="juliette"))
        self.client.delete(
            "EmailAddress", alias="juliette", address="juliette@logilab.fr"
        )
        self.assertFalse(self.client.find("EmailAddress", alias="juliette"))


class UtilsTests(unittest.TestCase):
    def test_base_url_slash_normalization(self):
        """test '/' can't appear twice in base_url

        e.g. we don't want ``http://www.logilab.org//``
        nor ``http://www.logilab.org//basepath``
        """
        proxy = CWProxy(base_url="http://www.logilab.org")
        self.assertEqual(proxy.base_url, ("http", "www.logilab.org"))
        proxy = CWProxy(base_url="http://www.logilab.org/")
        self.assertEqual(proxy.base_url, ("http", "www.logilab.org"))
        proxy = CWProxy(base_url="http://www.logilab.org/basepath")
        self.assertEqual(proxy.base_url, ("http", "www.logilab.org/basepath"))
        proxy = CWProxy(base_url="http://www.logilab.org/basepath/")
        self.assertEqual(proxy.base_url, ("http", "www.logilab.org/basepath"))

    def test_base_url_absolute_url(self):
        """test CWProxy.build_url parsing of absolute URLs"""
        proxy = CWProxy(base_url="http://www.logilab.org")
        for url in (
            "http://www.logilab.org",
            "http://www.logilab.org/path/to",
            "http://www.logilab.org/path/to/",
        ):
            self.assertEqual(url, proxy.build_url(url))

        proxy = CWProxy(base_url="http://www.logilab.org/with/path/")
        for url in (
            "http://www.logilab.org/with/path",
            "http://www.logilab.org/with/path/to/something",
            "http://www.logilab.org/with/path/to/something/",
        ):
            self.assertEqual(url, proxy.build_url(url))


if __name__ == "__main__":
    from unittest import main

    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

import sys

from . import cwproxy_for, get_config
from .cwproxy import CWProxy

import argparse

try:
    import argcomplete
except ImportError:
    argcomplete = None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "endpoint", nargs=1, choices=sorted(get_config()), help="endpoint"
    )
    parser.add_argument("path", nargs=1, help="path to query")
    parser.add_argument(
        "options",
        nargs="*",
        help=("options to be passed as query string " "(of the form key=value)"),
    )
    parser.add_argument(
        "-c",
        "--server-ca",
        dest="ca",
        default=None,
        help="Bundle CA to use to verify server certificate",
    )
    parser.add_argument(
        "-S",
        "--no-ssl",
        dest="verify",
        default=True,
        action="store_false",
        help=("do NOT verify ssl server certificate; ignored " "if --ca is given"),
    )
    parser.add_argument("-o", "--output", dest="output", default=None)
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        default=False,
        action="store_true",
        help=("Display HTTP request and response headers " "on non 200"),
    )

    if argcomplete is not None:
        argcomplete.autocomplete(parser)
    args = parser.parse_args()

    url = args.endpoint[0]
    path = args.path[0]
    verify = args.ca or args.verify  # path to the bundle CA or bool
    options = [tuple(arg.split("=", 1)) for arg in args.options]
    try:
        client = cwproxy_for(url, verify=verify)
    except ValueError:
        if url.startswith(("http://", "https://")):
            client = CWProxy(url, verify=verify)
        else:
            raise
    resp = client.get(path, options)
    if not args.output:
        out = sys.stdout
    else:
        from codecs import open

        out = open(args.output[0], "wb", encoding="utf-8", errors="ignore")
    if args.verbose:
        print("** Writing to {0}".format(out), file=sys.stderr)
    out.write(resp.text)
    if resp.status_code != 200:
        print(
            "** Request failed: {0} ({1})".format(resp.status_code, resp.text[:100]),
            file=sys.stderr,
        )
        if args.verbose:
            print("** Url: {0}".format(resp.request.url), file=sys.stderr),
            print("** Headers:", file=sys.stderr)
            print(
                "\n".join("   {0}: {1}".format(k, v) for k, v in resp.headers.items()),
                file=sys.stderr,
            )
            print("** Request Headers:", file=sys.stderr)
            print(
                "\n".join(
                    "   {0}: {1}".format(k, v) for k, v in resp.request.headers.items()
                ),
                file=sys.stderr,
            )


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# pylint: disable=W0142,W0403,W0404,W0613,W0622,W0622,W0704,R0904,C0103,E0611
#
# copyright 2003-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with CubicWeb.  If not, see <http://www.gnu.org/licenses/>.
"""Generic Setup script, takes package info from __pkginfo__.py file
"""
__docformat__ = "restructuredtext en"

from os.path import join, dirname

from setuptools import setup, find_packages
from codecs import open


# load metadata from the __pkginfo__.py file so there is no risk of conflict
# see https://packaging.python.org/en/latest/single_source_version.html
base_dir = dirname(__file__)

pkginfo = {}
with open(join(base_dir, "cwclientlib", "__pkginfo__.py")) as f:
    exec(f.read(), pkginfo)

# get required metadatas
modname = pkginfo["modname"]
version = pkginfo["version"]
license = pkginfo["license"]
description = pkginfo["description"]
web = pkginfo["web"]
author = pkginfo["author"]
author_email = pkginfo["author_email"]
classifiers = pkginfo["classifiers"]

with open(join(base_dir, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

# get optional metadatas
distname = pkginfo.get("distname", modname)
console_scripts = pkginfo.get("console_scripts", ())
include_dirs = pkginfo.get("include_dirs", ())
data_files = pkginfo.get("data_files")
ext_modules = pkginfo.get("ext_modules")
dependency_links = pkginfo.get("dependency_links", ())
install_requires = pkginfo.get("install_requires")
extras_require = pkginfo.get("extras_require")
test_suite = pkginfo.get("test_suite")


def install(**kwargs):
    """setup entry point"""
    return setup(
        name=distname,
        version=version,
        license=license,
        description=description,
        long_description=long_description,
        author=author,
        author_email=author_email,
        url=web,
        entry_points=dict(
            console_scripts=console_scripts,
        ),
        data_files=data_files,
        include_package_data=True,
        ext_modules=ext_modules,
        classifiers=classifiers,
        packages=find_packages(),
        install_requires=install_requires,
        extras_require=extras_require,
        dependency_links=dependency_links,
        test_suite=test_suite,
        python_requires=">=3.5",
        **kwargs
    )


if __name__ == "__main__":
    install()



# -*- coding: utf-8 -*-
#
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

"""
Build common queries usable with rqlcontroller

Each returned query is a couple (rql, kwargs) to be used to build the
'queries' argument of the CWProxy.rqlio() method, like::


  import cwproxy, builders

  client = cwproxy.CWProxy('http://www.cubicweb.org/')
  queries = [builders.create_entity('CWUser', login='Babar', password='pwd'),
             builders.build_trinfo('__r0', 'disable', 'not yet activated'),
            ]
  resp = client.rqlio(queries)


"""


def create_entity(cwetype, **kw):
    """Build the rqlio query thath will create a new <cwetype> entity,
    with given attributes.
    """
    args = ", ".join("X %s %%(%s)s" % (key, key) for key in kw)
    return ("INSERT %s X: %s" % (cwetype, args), kw)


def build_trinfo(eid, transition_name, comment=None):
    """Build the rqlio query that will fire the transistion <transition_name>
    for the entity <eid> and attach the optional <comment>.
    """
    return (
        "INSERT TrInfo X: X comment %(comment)s, "
        "X by_transition BT, X wf_info_for Y "
        "WHERE Y eid %(eid)s, Y in_state S, S state_of W, "
        "BT transition_of W, BT name %(trname)s",
        {"eid": eid, "trname": transition_name, "comment": comment or ""},
    )


def build_multiple_trinfo(eids, transition_name, comment=None):
    """Build the rqlio query that will fire the transistion <transition_name>
    for several entities in <eids> and attach the optional <comment>.
    """
    return (
        "INSERT TrInfo X: X comment %(comment)s, "
        "X by_transition BT, X wf_info_for Y "
        "WHERE Y eid in ({0}), Y in_state S, S state_of W, "
        "BT transition_of W, BT name %(trname)s".format(
            ",".join(str(eid) for eid in eids)
        ),
        {
            "trname": transition_name,
            "comment": comment or "",
        },
    )



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with cwclientlib. If not, see <http://www.gnu.org/licenses/>.

from code import interact

from .cwcompleter import setup_autocompleter
from . import cwproxy_for, get_config
from .cwproxy import CWProxy

import argparse

try:
    import argcomplete
except ImportError:
    argcomplete = None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c",
        "--server-ca",
        dest="ca",
        default=None,
        help="Bundle CA to use to verify server certificate",
    )
    parser.add_argument(
        "endpoint", nargs=1, choices=sorted(get_config()), help="endpoint"
    )
    parser.add_argument("query", nargs="*", help="rql query")
    if argcomplete is not None:
        argcomplete.autocomplete(parser)
    args = parser.parse_args()

    url = args.endpoint[0]
    try:
        client = cwproxy_for(url, verify=args.ca)
    except ValueError:
        if url.startswith(("http://", "https://")):
            client = CWProxy(url, verify=args.ca)
        else:
            raise
    namespace = {"client": client, "rql": client.execute, "rqlio": client.rqlio}
    setup_autocompleter(client, namespace=namespace)
    if not args.query:
        interact(
            "You are connected to {0}".format(client.build_url("")), local=namespace
        )
    elif len(args.query) == 1:
        client.execute(args.query[0])
    else:
        for query in args.query:
            print(">>>", query)
            client.execute(query)


if __name__ == "__main__":
    main()



# pylint: disable=W0622
# copyright 2014-2017 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of cwclientlib.
#
# cwclientlib is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 2.1 of the License, or (at your
# option) any later version.
#
# cwclientlib is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License
# for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with cwclientlib. If not, see <https://www.gnu.org/licenses/>.

"""cwclientlib application packaging information"""

import sys

modname = "cwclientlib"
distname = "cwclientlib"

numversion = (1, 1, 0)
version = ".".join(str(num) for num in numversion)

license = "LGPL"
author = "LOGILAB S.A. (Paris, FRANCE)"
author_email = "contact@logilab.fr"
description = "A Python library to easily build CubicWeb clients"
web = "https://www.cubicweb.org/project/%s" % distname

cliversion = ""
if sys.version_info[0] > 2:
    cliversion = "3"
console_scripts = [
    "cwrql{0}=cwclientlib.cwrql:main".format(cliversion),
    "cwget{0}=cwclientlib.cwget:main".format(cliversion),
    "cwshell{0}=cwclientlib.cwshell:main".format(cliversion),
]

install_requires = ["requests >= 2"]
extras_require = {
    "yaml": ["PyYAML"],
}
test_suite = "cwclientlib.test"
tests_require = [
    "cubicweb >= 3.27",
    "cubicweb-signedrequest",
    "cubicweb-rqlcontroller",
    "cubicweb-file",
    "twisted",
    "pyramid",
    "PyYAML",
    "requests",
]

classifiers = [
    "Environment :: Web Environment",
    "Framework :: CubicWeb",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
]
